"""
Module for parsing FTP data.

Currently, this module contains classes for parsing FTP ``LIST`` command
output from a variety of FTP servers. In the future, this module may be
extended to handle other FTP parsing chores. (Or not.)

The FTP ``LIST`` parsing logic was adapted for Python from D. J. Bernstein's
``ftpparse.c`` library. See http://cr.yp.to/ftpparse.html. The logic in this
module is functionally similar to Bernstein's parser, with the following
differences:

    - Bernstein's C-specific logic has been made more Python-like.
    - The basic parser is encapsulated inside an `FTPListDataParser` class,
      instead of a function.
    - The ``ID_TYPE`` and ``MTIME_TYPE`` values are enumerations.
    - ``SIZE_TYPE`` is not supported (since it was always being set to the
      same value anyway).

Currently covered formats:

    - `EPLF`_
    - UNIX *ls*, with or without group ID
    - Microsoft FTP Service
    - Windows NT FTP Server
    - VMS
    - WFTPD
    - NetPresenz (Mac)
    - NetWare
    - MSDOS

.. _EPLF: http://cr.yp.to/ftp/list/eplf.html

Definitely not covered:

    - Long VMS filenames, with information split across two lines.
    - NCSA Telnet FTP server. Has LIST = NLST (and bad NLST for directories).
"""

__docformat__ = 'restructuredtext en'

# ---------------------------------------------------------------------------
# Globals
# ---------------------------------------------------------------------------

import time
from enum import Enum

# ---------------------------------------------------------------------------
# Exports
# ---------------------------------------------------------------------------

__all__ = ['parse_ftp_list_line',
           'FTPListData',
           'FTPListDataParser',
           'ID_TYPE',
           'MTIME_TYPE']

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

MONTHS = ('jan', 'feb', 'mar', 'apr', 'may', 'jun',
          'jul', 'aug', 'sep', 'oct', 'nov', 'dec')

MTIME_TYPE = Enum('UNKNOWN', 'LOCAL', 'REMOTE_MINUTE', 'REMOTE_DAY')
"""
``MTIME_TYPE`` identifies how a modification time ought to be interpreted
(assuming the caller cares).

    - ``LOCAL``: Time is local to the client, granular to (at least) the minute
    - ``REMOTE_MINUTE``: Time is local to the server and granular to the minute
    - ``REMOTE_DAY``: Time is local to the server and granular to the day.
    - ``UNKNOWN``: Time's locale is unknown.
"""

ID_TYPE = Enum('UNKNOWN', 'FULL')
"""
``ID_TYPE`` identifies how a file's identifier should be interpreted.

    - ``FULL``: The ID is known to be complete.
    - ``UNKNOWN``: The ID is not set or its type is unknown.
"""

# ---------------------------------------------------------------------------
# Globals
# ---------------------------------------------------------------------------

now = time.time()
current_year = time.localtime().tm_year

# ---------------------------------------------------------------------------
# Classes
# ---------------------------------------------------------------------------

class FTPListData(object):
    """
    The `FTPListDataParser` class's ``parse_line()`` method returns an
    instance of this class, capturing the parsed data.

    :IVariables:
        name : str
            The name of the file, if parsable
        try_cwd : bool
            ``True`` if the entry might be a directory (i.e., the caller
            might want to try an FTP ``CWD`` command), ``False`` if it
            cannot possibly be a directory.
        try_retr : bool
            ``True`` if the entry might be a retrievable file (i.e., the caller
            might want to try an FTP ``RETR`` command), ``False`` if it
            cannot possibly be a file.
        size : long
            The file's size, in bytes
        mtime : long
            The file's modification time, as a value that can be passed to
            ``time.localtime()``.
        mtime_type : `MTIME_TYPE`
            How to interpret the modification time. See `MTIME_TYPE`.
        id : str
            A unique identifier for the file. The unique identifier is unique
            on the *server*. On a Unix system, this identifier might be the
            device number and the file's inode; on other system's, it might
            be something else. It's also possible for this field to be ``None``.
        id_type : `ID_TYPE`
            How to interpret the identifier. See `ID_TYPE`.
   """

    def __init__(self, raw_line):
        self.raw_line = raw_line
        self.name = None
        self.try_cwd = False
        self.try_retr = False
        self.size = 0
        self.mtime_type = MTIME_TYPE.UNKNOWN
        self.mtime = 0
        self.id_type = ID_TYPE.UNKNOWN
        self.id = None

class FTPListDataParser(object):
    """
    An ``FTPListDataParser`` object can be used to parse one or more lines
    that were retrieved by an FTP ``LIST`` command that was sent to a remote
    server.
    """
    def __init__(self):
        pass

    def parse_line(self, ftp_list_line):
        """
        Parse a line from an FTP ``LIST`` command.

        :Parameters:
            ftp_list_line : str
                The line of output

        :rtype: `FTPListData`
        :return: An `FTPListData` object describing the parsed line, or
                 ``None`` if the line could not be parsed. Note that it's
                 possible for this method to return a partially-filled
                 `FTPListData` object (e.g., one without a name).
        """
        buf = ftp_list_line

        if len(buf) < 2: # an empty name in EPLF, with no info, could be 2 chars
            return None

        c = buf[0]
        if c == '+':
            return self._parse_EPLF(buf)

        elif c in 'bcdlps-':
            return self._parse_unix_style(buf)

        i = buf.find(';')
        if i > 0:
            return self._parse_multinet(buf, i)

        if c in '0123456789':
            return self._parse_msdos(buf)

        return None

    # UNIX ls does not show the year for dates in the last six months.
    # So we have to guess the year.
    #
    # Apparently NetWare uses ``twelve months'' instead of ``six months''; ugh.
    # Some versions of ls also fail to show the year for future dates.

    def _guess_time(self, month, mday, hour=0, minute=0):
        year = None
        t = None

        for year in range(current_year - 1, current_year + 100):
            t = self._get_mtime(year, month, mday, hour, minute)
            if (now - t) < (350 * 86400):
                return t

        return 0

    def _get_mtime(self, year, month, mday, hour=0, minute=0, second=0):
        return time.mktime((year, month, mday, hour, minute, second, 0, 0, -1))

    def _get_month(self, buf):
        if len(buf) == 3:
            for i in range(0, 12):
                if buf.lower().startswith(MONTHS[i]):
                    return i+1
        return -1

    def _parse_EPLF(self, buf):
        result = FTPListData(buf)

        # see http://cr.yp.to/ftp/list/eplf.html
        #"+i8388621.29609,m824255902,/,\tdev"
        #"+i8388621.44468,m839956783,r,s10376,\tRFCEPLF"
        i = 1
        for j in range(1, len(buf)):
            if buf[j] == '\t':
                result.name = buf[j+1:]
                break

            if buf[j] == ',':
                c = buf[i]
                if c == '/':
                    result.try_cwd = True
                elif c == 'r':
                    result.try_retr = True
                elif c == 's':
                    result.size = long(buf[i+1:j])
                elif c == 'm':
                    result.mtime_type = MTIME_TYPE.LOCAL
                    result.mtime = long(buf[i+1:j])
                elif c == 'i':
                    result.id_type = ID_TYPE.FULL
                    result.id = buf[i+1:j-i-1]

                i = j + 1

        return result

    def _parse_unix_style(self, buf):
        # UNIX-style listing, without inum and without blocks:
        # "-rw-r--r--   1 root     other        531 Jan 29 03:26 README"
        # "dr-xr-xr-x   2 root     other        512 Apr  8  1994 etc"
        # "dr-xr-xr-x   2 root     512 Apr  8  1994 etc"
        # "lrwxrwxrwx   1 root     other          7 Jan 25 00:17 bin -> usr/bin"
        #
        # Also produced by Microsoft's FTP servers for Windows:
        # "----------   1 owner    group         1803128 Jul 10 10:18 ls-lR.Z"
        # "d---------   1 owner    group               0 May  9 19:45 Softlib"
        #
        # Also WFTPD for MSDOS:
        # "-rwxrwxrwx   1 noone    nogroup      322 Aug 19  1996 message.ftp"
        #
        # Also NetWare:
        # "d [R----F--] supervisor            512       Jan 16 18:53    login"
        # "- [R----F--] rhesus             214059       Oct 20 15:27    cx.exe"
        #
        # Also NetPresenz for the Mac:
        # "-------r--         326  1391972  1392298 Nov 22  1995 MegaPhone.sit"
        # "drwxrwxr-x               folder        2 May 10  1996 network"

        result = FTPListData(buf)

        buflen = len(buf)
        c = buf[0]
        if c == 'd':
            result.try_cwd = True
        if c == '-':
            result.try_retr = True
        if c == 'l':
            result.try_retr = True
            result.try_cwd = True

        state = 1
        i = 0
        tokens = buf.split()
        for j in range(1, buflen):
            if (buf[j] == ' ') and (buf[j - 1] != ' '):
                if state == 1:  # skipping perm
                    state = 2

                elif state == 2: # skipping nlink
                    state = 3
                    if ((j - i) == 6) and (buf[i] == 'f'): # NetPresenz
                        state = 4

                elif state == 3: # skipping UID/GID
                    state = 4

                elif state == 4: # getting tentative size
                    try:
                        size = long(buf[i:j])
                    except ValueError:
                        pass
                    state = 5

                elif state == 5: # searching for month, else getting tentative size
                    month = self._get_month(buf[i:j])
                    if month >= 0:
                        state = 6
                    else:
                        size = long(buf[i:j])

                elif state == 6: # have size and month
                    mday = long(buf[i:j])
                    state = 7

                elif state == 7: # have size, month, mday
                    if (j - i == 4) and (buf[i+1] == ':'):
                        hour = long(buf[i])
                        minute = long(buf[i+2:i+4])
                        result.mtime_type = MTIME_TYPE.REMOTE_MINUTE
                        result.mtime = self._guess_time(month, mday, hour, minute)
                    elif (j - i == 5) and (buf[i+2] == ':'):
                        hour = long(buf[i:i+2])
                        minute = long(buf[i+3:i+5])
                        result.mtime_type = MTIME_TYPE.REMOTE_MINUTE
                        result.mtime = self._guess_time(month, mday, hour, minute)
                    elif j - i >= 4:
                        year = long(buf[i:j])
                        result.mtimetype = MTIME_TYPE.REMOTE_DAY
                        result.mtime = self._get_mtime(year, month, mday)
                    else:
                        break

                    result.name = buf[j+1:]
                    state = 8
                elif state == 8: # twiddling thumbs
                    pass

                i = j + 1
                while (i < buflen) and (buf[i] == ' '):
                    i += 1

        #if state != 8:
            #return None

        result.size = size

        if c == 'l':
            i = 0
            while (i + 3) < len(result.name):
                if result.name[i:i+4] == ' -> ':
                    result.name = result.name[:i]
                    break
                i += 1

        # eliminate extra NetWare spaces
        if (buf[1] == ' ') or (buf[1] == '['):
            namelen = len(result.name)
            if namelen > 3:
                result.name = result.name.strip()

        return result

    def _parse_multinet(self, buf, i):

        # MultiNet (some spaces removed from examples)
        # "00README.TXT;1      2 30-DEC-1996 17:44 [SYSTEM] (RWED,RWED,RE,RE)"
        # "CORE.DIR;1          1  8-SEP-1996 16:09 [SYSTEM] (RWE,RWE,RE,RE)"
        # and non-MultiNet VMS:
        #"CII-MANUAL.TEX;1  213/216  29-JAN-1996 03:33:12  [ANONYMOU,ANONYMOUS]   (RWED,RWED,,)"

        result = FTPListData(buf)
        result.name = buf[:i]
        buflen = len(buf)

        if i > 4:
            if buf[i-4:i] == '.DIR':
                result.name = result.name[0:-4]
                result.try_cwd = True

        if not result.try_cwd:
            result.try_retr = True

        try:
            i = buf.index(' ', i)
            i = _skip(buf, i, ' ')
            i = buf.index(' ', i)
            i = _skip(buf, i, ' ')

            j = i

            j = buf.index('-', j)
            mday = long(buf[i:j])

            j = _skip(buf, j, '-')
            i = j
            j = buf.index('-', j)
            month = self._get_month(buf[i:j])
            if month < 0:
                raise IndexError

            j = _skip(buf, j, '-')
            i = j
            j = buf.index(' ', j)
            year = long(buf[i:j])

            j = _skip(buf, j, ' ')
            i = j

            j = buf.index(':', j)
            hour = long(buf[i:j])
            j = _skip(buf, j, ':')
            i = j

            while (buf[j] != ':') and (buf[j] != ' '):
                j += 1
                if j == buflen:
                    raise IndexError # abort, abort!

            minute = long(buf[i:j])

            result.mtimetype = MTIME_TYPE.REMOTE_MINUTE
            result.mtime = self._get_mtime(year, month, mday, hour, minute)

        except IndexError:
            pass

        return result

    def _parse_msdos(self, buf):
        # MSDOS format
        # 04-27-00  09:09PM       <DIR>          licensed
        # 07-18-00  10:16AM       <DIR>          pub
        # 04-14-00  03:47PM                  589 readme.htm

        buflen = len(buf)
        i = 0
        j = 0

        try:
            result = FTPListData(buf)

            j = buf.index('-', j)
            month = long(buf[i:j])

            j = _skip(buf, j, '-')
            i = j
            j = buf.index('-', j)
            mday = long(buf[i:j])

            j = _skip(buf, j, '-')
            i = j
            j = buf.index(' ', j)
            year = long(buf[i:j])
            if year < 50:
                year += 2000
            if year < 1000:
                year += 1900

            j = _skip(buf, j, ' ')
            i = j
            j = buf.index(':', j)
            hour = long(buf[i:j])
            j = _skip(buf, j, ':')
            i = j
            while not (buf[j] in 'AP'):
                j += 1
                if j == buflen:
                    raise IndexError
            minute = long(buf[i:j])

            if buf[j] == 'A':
                j += 1
                if j == buflen:
                    raise IndexError

            if buf[j] == 'P':
                hour = (hour + 12) % 24
                j += 1
                if j == buflen:
                    raise IndexError

            if buf[j] == 'M':
                j += 1
                if j == buflen:
                    raise IndexError

            j = _skip(buf, j, ' ')
            if buf[j] == '<':
                result.try_cwd = True
                j = buf.index(' ', j)
            else:
                i = j
                j = buf.index(' ', j)

                result.size = long(buf[i:j])
                result.try_retr = True

            j = _skip(buf, j, ' ')

            result.name = buf[j:]
            result.mtimetype = MTIME_TYPE.REMOTE_MINUTE
            result.mtime = self._get_mtime(year, month, mday, hour, minute)
        except IndexError:
            pass

        return result


# ---------------------------------------------------------------------------
# Public Functions
# ---------------------------------------------------------------------------

def parse_ftp_list_line(ftp_list_line):
    """
    Convenience function that instantiates an `FTPListDataParser` object
    and passes ``ftp_list_line`` to the object's ``parse_line()`` method,
    returning the result.

    :Parameters:
        ftp_list_line : str
            The line of output

    :rtype: `FTPListData`
    :return: An `FTPListData` object describing the parsed line, or
             ``None`` if the line could not be parsed. Note that it's
             possible for this method to return a partially-filled
             `FTPListData` object (e.g., one without a name).
    """
    return FTPListDataParser().parse_line(ftp_list_line)

# ---------------------------------------------------------------------------
# Private Functions
# ---------------------------------------------------------------------------

def _skip(s, i, c):
    while s[i] == c:
        i += 1
        if i == len(s):
            raise IndexError
    return i

# ---------------------------------------------------------------------------
# Main (Tester)
# ---------------------------------------------------------------------------

if __name__ == '__main__':
    test_data = [
        # ELPF

        {'line':    '+i9872342.32142,m1229473595,/,\tpub',
         'type':    'ELPF',
         'size':    0,
         'time':    (2008, 12, 16, 19, 26, 35, 0, 0, -1),
         'name':    'pub',
         'try_cwd': True},

        {'line':    '+i9872342.32142,m1229473595,r,s10376,\tREADME.txt',
         'type':    'ELPF',
         'size':    10376,
         'time':    (2008, 12, 16, 19, 26, 35, 0, 0, -1),
         'name':    'README.txt',
         'try_cwd': False},

        # Unix

        {'line':    '-rw-r--r--   1 root     other     531 Jan 29 03:26 README',
         'type':    'Unix',
         'size':    531,
         'time':    (current_year, 1, 29, 03, 26, 0, 0, 0, -1),
         'name':    'README',
         'try_cwd': False},

        {'line':    'dr-xr-xr-x   2 root     other        512 Apr  8  2003 etc',
         'type':    'Unix',
         'size':    512,
         'time':    (2003, 4, 8, 0, 0, 0, 0, 0, -1),
         'name':    'etc',
         'try_cwd': True},

        {'line':    '-rw-r--r--   1 1356107  15000      4356349 Nov 23 11:34 09 Ribbons Undone.wma',
         'type':    'Unix',
         'size':    4356349,
         'time':    (current_year, 11, 23, 11, 34, 0, 0, 0, -1),
         'name':    '09 Ribbons Undone.wma',
         'try_cwd': False},

        # Microsoft Windows

        {'line':    '----------   1 owner    group         1803128 Jul 10 10:18 ls-lR.Z',
         'type':    'Windows',
         'size':    1803128,
         'time':    (current_year, 7, 10, 10, 18, 0, 0, 0, -1),
         'name':    'ls-lR.Z',
         'try_cwd': False},

        {'line':    'd---------   1 owner    group               0 May  9 19:45 foo bar',
         'type':    'Windows',
         'size':    0,
         'time':    (current_year, 5, 9, 19, 45, 0, 0, 0, -1),
         'name':    'foo bar',
         'try_cwd': True},

        # NetWare

        {'line':    'd [R----F--] supervisor    512    Jan 16 18:53    login',
         'type':    'NetWare',
         'size':    512,
         'time':    (current_year, 1, 16, 18, 53, 0, 0, 0, -1),
         'name':    'login',
         'try_cwd': True},

        # NetPresenz

        {'line':    'drwxrwxr-x               folder   2 May 10  1996 bar.sit',
         'type':    'NetPresenz/Mac',
         'size':    2,
         'time':    (1996, 5, 10, 0, 0, 0, 0, 0, -1),
         'name':    'bar.sit',
         'try_cwd': True},

        # MultiNet/VMS (no size with these)

        {'line':    'CORE.DIR;1      1 8-NOV-1999 07:02 [SYSTEM] (RWED,RWED,RE,RE)',
         'type':    'MultiNet/VMS',
         'size':    0,
         'time':    (1999, 11, 8, 7, 2, 0, 0, 0, -1),
         'name':    'CORE',
         'try_cwd': True},

        {'line':    '00README.TXT;1      2 30-DEC-1976 17:44 [SYSTEM] (RWED,RWED,RE,RE)',
         'type':    'MultiNet/VMS',
         'size':    0,
         'time':    (1976, 12, 30, 17, 44, 0, 0, 0, -1),
         'name':    '00README.TXT',
         'try_cwd': False},

        {'line':    'CII-MANUAL.TEX;1  213/216  29-JAN-1996 03:33:12  [ANONYMOU,ANONYMOUS]   (RWED,RWED,,)',
         'type':    'MultiNet/VMS',
         'size':    0,
         # Doesn't parse the seconds
         'time':    (1996, 1, 29, 03, 33, 0, 0, 0, -1),
         'name':    'CII-MANUAL.TEX',
         'try_cwd': False},

        # MS-DOS

        {'line':    '04-27-00  09:09PM       <DIR>          licensed',
         'type':    'MS-DOS',
         'size':    0,
         'time':    (2000, 4, 27, 21, 9, 0, 0, 0, -1),
         'name':    'licensed',
         'try_cwd': True},

        {'line':    '11-18-03  10:16AM       <DIR>          pub',
         'type':    'MS-DOS',
         'size':    0,
         'time':    (2003, 11, 18, 10, 16, 0, 0, 0, -1),
         'name':    'pub',
         'try_cwd': True},

        {'line':    '04-14-99  03:47PM                  589 readme.htm',
         'type':    'MS-DOS',
         'size':    589,
         'time':    (1999, 04, 14, 15, 47, 0, 0, 0, -1),
         'name':    'readme.htm',
         'try_cwd': False},
    ]

    def assertEquals(test_value, expected_value, prefix=None):
        error_message = '%s: ' % prefix if prefix else ''
        error_message += 'Expected %s, got %s' % (expected_value, test_value)
        assert test_value == expected_value, error_message


    parser = FTPListDataParser()
    i = 0
    for test in test_data:
        line = test['line']
        prefix = 'Test %d (%s)' % (i, test['type'])
        print '%s: "%s"' % (prefix, test['name'])
        result = parser.parse_line(line)
        assertEquals(result.raw_line, line, prefix)
        assertEquals(result.size, test['size'], prefix)
        assertEquals(result.name, test['name'], prefix)
        assertEquals(result.try_cwd, test['try_cwd'], prefix)
        expected_time = time.mktime(test['time'])
        assertEquals(time.localtime(result.mtime),
                     time.localtime(expected_time),
                     prefix)
        assertEquals(result.mtime, expected_time, prefix)
        i += 1




# $Id: a619749fa4d80c458a3eefe2b419922edc0e16a8 $

"""
SQLite3 extended database driver.
"""

__docformat__ = "restructuredtext en"

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

import os
import sys
import re

from grizzled.db.base import (DBDriver, Error, Warning, TableMetadata,
                              IndexMetadata, RDBMSMetadata)

# ---------------------------------------------------------------------------
# Exports
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# Classes
# ---------------------------------------------------------------------------

class SQLite3Driver(DBDriver):
    """DB Driver for SQLite, using the pysqlite DB API module."""

    def get_import(self):
        import sqlite3
        return sqlite3

    def get_display_name(self):
        return "SQLite3"

    def do_connect(self,
                   host=None,
                   port=None,
                   user='',
                   password='',
                   database='default'):
        dbi = self.get_import()
        return dbi.connect(database=database, isolation_level=None)

    def get_rdbms_metadata(self, cursor):
        import sqlite3
        return RDBMSMetadata('SQLite', 'SQLite 3', sqlite3.sqlite_version)

    def get_tables(self, cursor):
        cursor.execute("select name from sqlite_master where type = 'table'")
        table_names = []
        rs = cursor.fetchone()
        while rs is not None:
            table_names += [rs[0]]
            rs = cursor.fetchone()

        return table_names

    def get_table_metadata(self, table, cursor):
        self._ensure_valid_table(cursor, table)

        # The table_info pragma returns results looking something like this:
        #
        # cid name            type              notnull dflt_value pk
        # --- --------------- ----------------- ------- ---------- --
        # 0   id              integer           99      NULL       1
        # 1   action_time     datetime          99      NULL       0
        # 2   user_id         integer           99      NULL       0
        # 3   content_type_id integer           0       NULL       0
        # 4   object_id       text              0       NULL       0
        # 5   object_repr     varchar(200)      99      NULL       0
        # 6   action_flag     smallint unsigned 99      NULL       0
        # 7   change_message  text              99      NULL       0

        cursor.execute('PRAGMA table_info(%s)' % table)
        rs = cursor.fetchone()
        result = []

        char_field_re = re.compile(r'(varchar|char)\((\d+)\)')
        while rs is not None:
            (id, name, type, not_null, default_value, is_primary) = rs
            m = char_field_re.match(type)
            if m:
                type = m.group(1)
                try:
                    max_char_size = int(m.group(2))
                except ValueError:
                    log.error('Bad number in "%s" type for column "%s"' %
                              (type, name))
            else:
                max_char_size = 0

            data = TableMetadata(name, type, max_char_size, 0, 0, not not_null)
            result += [data]

            rs = cursor.fetchone()

        return result

    def get_index_metadata(self, table, cursor):
        self._ensure_valid_table(cursor, table)

        # First, get the list of indexes for the table, using the appropriate
        # pragma. The pragma returns output like this:
        #
        # seq name    unique
        # --- ------- ------
        # 0   id        0
        # 1   name      0
        # 2   address   0

        result = []

        cursor.execute("PRAGMA index_list('%s')" % table)
        indexes = []
        rs = cursor.fetchone()
        while rs is not None:
            indexes += [(rs[1], rs[2])]
            rs = cursor.fetchone()

        # Now, get the data about each index, using another pragma. This
        # pragma returns data like this:
        #
        # seqno cid name
        # ----- --- ---------------
        # 0     3   content_type_id

        for name, unique in indexes:
            cursor.execute("PRAGMA index_info('%s')" % name)
            rs = cursor.fetchone()
            columns = []
            while rs is not None:
                columns += [rs[2]]
                rs = cursor.fetchone()

            description = 'UNIQUE' if unique else ''
            result += [IndexMetadata(name, columns, description)]

        return result



#!/usr/bin/env python

# NOTE: Documentation is intended to be processed by epydoc and contains
# epydoc markup.

"""
Overview
========

The ``grizzled.os`` module contains some operating system-related functions and
classes. It is a conceptual extension of the standard Python ``os`` module.
"""

from __future__ import absolute_import

__docformat__ = "restructuredtext en"

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

import logging
import os as _os
import sys
import glob
from contextlib import contextmanager

from grizzled.decorators import deprecated

# ---------------------------------------------------------------------------
# Exports
# ---------------------------------------------------------------------------

__all__ = ['daemonize', 'DaemonError', 'working_directory',
           'file_separator', 'path_separator']

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

# Default daemon parameters.
# File mode creation mask of the daemon.
UMASK = 0

# Default working directory for the daemon.
WORKDIR = "/"

# Default maximum for the number of available file descriptors.
MAXFD = 1024

# The standard I/O file descriptors are redirected to /dev/null by default.
if (hasattr(_os, "devnull")):
    NULL_DEVICE = _os.devnull
else:
    NULL_DEVICE = "/dev/null"

# The path separator for the operating system.

PATH_SEPARATOR = {'nt' : ';', 'posix' : ':'}
FILE_SEPARATOR = {'nt' : '\\', 'posix' : '/'}

# ---------------------------------------------------------------------------
# Logging
# ---------------------------------------------------------------------------

log = logging.getLogger('grizzled.os')

# ---------------------------------------------------------------------------
# Public classes
# ---------------------------------------------------------------------------

class DaemonError(OSError):
    """
    Thrown by ``daemonize()`` when an error occurs while attempting to create
    a daemon.
    """
    pass

# ---------------------------------------------------------------------------
# Public functions
# ---------------------------------------------------------------------------

def path_separator():
    """
    Get the path separator for the current operating system. The path
    separator is used to separate elements of a path string, such as
    "PATH" or "CLASSPATH". (It's a ":" on Unix-like systems and a ";"
    on Windows.)

    :rtype: str
    :return: the path separator
    """
    return PATH_SEPARATOR[_os.name]

@deprecated(since='0.7', message='Use os.path.sep, instead.')
def file_separator():
    """
    Get the file separator for the current operating system. The file
    separator is used to separate file elements in a pathname. (It's
    "/" on Unix-like systems and a "\\\\" on Windows.)
    
    **Deprecated**. Use the standard Python ``os.path.sep`` variable,
    instead.

    :rtype: str
    :return: the file separator
    """
    return FILE_SEPARATOR[_os.name]

def path_elements(path):
    """
    Given a path string value (e.g., the value of the environment variable
    ``PATH``), this generator function yields each item in the path.

    :Parameters:
        path
            the path to break up
    """
    for p in path.split(path_separator()):
        yield p

@contextmanager
def working_directory(directory):
    """
    This function is intended to be used as a ``with`` statement context
    manager. It allows you to replace code like this:

    .. python::

        original_directory = _os.getcwd()
        try:
            _os.chdir(some_dir)
            ... bunch of code ...
        finally:
            _os.chdir(original_directory)

    with something simpler:

    .. python ::

        from __future__ import with_statement
        from grizzled.os import working_directory

        with working_directory(some_dir):
            ... bunch of code ...

    :Parameters:
        directory : str
            directory in which to execute

    :return: yields the ``directory`` parameter
    """
    original_directory = _os.getcwd()
    try:
        _os.chdir(directory)
        yield directory

    finally:
        _os.chdir(original_directory)

def find_command(command_name, path=None):
    """
    Determine whether the specified system command exists in the specified
    path.

    :Parameters:
        command_name
            The (simple) filename of the command to find. May be a glob
            string.

        path
            The path to search, as a list or a string. If this parameter
            is a string, then it is split using the operating system-specific
            path separator. If this parameter is missing, then the ``PATH``
            environment variable is used

    :rtype: str
    :return: The path to the first command that matches ``command_name``, or 
             ``None`` if not found
    """
    if not path:
        path = _os.environ.get('PATH', '.')

    if type(path) == str:
        path = path.split(path_separator())
    elif type(path) == list:
        pass
    else:
        assert False, 'path parameter must be a list or a string'

    found = None
    for p in path:
        full_path = _os.path.join(p, command_name)
        for p2 in glob.glob(full_path):
            if _os.access(p2, _os.X_OK):
                found = p2
                break

        if found:
            break

    return found

def spawnd(path, args, pidfile=None):
    """
    Run a command as a daemon. This method is really just shorthand for the
    following code:
    
    .. python::
    
        daemonize(pidfile=pidfile)
        _os.execv(path, args)

    :Parameters:
        path : str
            Full path to program to run
            
        args : list
            List of command arguments. The first element in this list must
            be the command name (i.e., arg0).
            
        pidfile : str
            Path to file to which to write daemon's process ID. The string may
            contain a ``${pid}`` token, which is replaced with the process ID
            of the daemon. e.g.: ``/var/run/myserver-${pid}``
    """
    daemonize(no_close=True, pidfile=pidfile)
    _os.execv(path, args)

def daemonize(no_close=False, pidfile=None):
    """
    Convert the calling process into a daemon. To make the current Python
    process into a daemon process, you need two lines of code:

    .. python::

        from grizzled.os import daemonize
        daemonize.daemonize()

    If ``daemonize()`` fails for any reason, it throws a ``DaemonError``,
    which is a subclass of the standard ``OSError`` exception. also logs debug
    messages, using the standard Python ``logging`` package, to channel
    "grizzled.os.daemon".

    **Adapted from:** http://www.clapper.org/software/daemonize/

    **See Also:**

    - Stevens, W. Richard. *Unix Network Programming* (Addison-Wesley, 1990).

    :Parameters:
        no_close : bool
            If ``True``, don't close the file descriptors. Useful if the
            calling process has already redirected file descriptors to an
            output file. **Warning**: Only set this parameter to ``True`` if
            you're *sure* there are no open file descriptors to the calling
            terminal. Otherwise, you'll risk having the daemon re-acquire a
            control terminal, which can cause it to be killed if someone logs
            off that terminal.

        pidfile : str
            Path to file to which to write daemon's process ID. The string may
            contain a ``${pid}`` token, which is replaced with the process ID
            of the daemon. e.g.: ``/var/run/myserver-${pid}``

    :raise DaemonError: Error during daemonizing
    """
    log = logging.getLogger('grizzled.os.daemon')

    def __fork():
        try:
            return _os.fork()
        except OSError, e:
            raise DaemonError, ('Cannot fork', e.errno, e.strerror)

    def __redirect_file_descriptors():
        import resource  # POSIX resource information
        maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        if maxfd == resource.RLIM_INFINITY:
            maxfd = MAXFD

        # Close all file descriptors.

        for fd in range(0, maxfd):
            # Only close TTYs.
            try:
                _os.ttyname(fd)
            except:
                continue

            try:
                _os.close(fd)
            except OSError:
                # File descriptor wasn't open. Ignore.
                pass

            # Redirect standard input, output and error to something safe.
            # os.open() is guaranteed to return the lowest available file
            # descriptor (0, or standard input). Then, we can dup that
            # descriptor for standard output and standard error.

            _os.open(NULL_DEVICE, _os.O_RDWR)
            _os.dup2(0, 1)
            _os.dup2(0, 2)


    if _os.name != 'posix':
        import errno
        raise DaemonError, \
              ('daemonize() is only supported on Posix-compliant systems.',
               errno.ENOSYS, _os.strerror(errno.ENOSYS))

    try:
        # Fork once to go into the background.

        log.debug('Forking first child.')
        pid = __fork()
        if pid != 0:
            # Parent. Exit using os._exit(), which doesn't fire any atexit
            # functions.
            _os._exit(0)

        # First child. Create a new session. os.setsid() creates the session
        # and makes this (child) process the process group leader. The process
        # is guaranteed not to have a control terminal.
        log.debug('Creating new session')
        _os.setsid()

        # Fork a second child to ensure that the daemon never reacquires
        # a control terminal.
        log.debug('Forking second child.')
        pid = __fork()
        if pid != 0:
            # Original child. Exit.
            _os._exit(0)

        # This is the second child. Set the umask.
        log.debug('Setting umask')
        _os.umask(UMASK)

        # Go to a neutral corner (i.e., the primary file system, so
        # the daemon doesn't prevent some other file system from being
        # unmounted).
        log.debug('Changing working directory to "%s"' % WORKDIR)
        _os.chdir(WORKDIR)

        # Unless no_close was specified, close all file descriptors.
        if not no_close:
            log.debug('Redirecting file descriptors')
            __redirect_file_descriptors()

        if pidfile:
            from string import Template
            t = Template(pidfile)
            pidfile = t.safe_substitute(pid=str(_os.getpid()))
            open(pidfile, 'w').write(str(_os.getpid()) + '\n')

    except DaemonError:
        raise

    except OSError, e:
        raise DaemonError, ('Unable to daemonize()', e.errno, e.strerror)

# ---------------------------------------------------------------------------
# Main program (for testing)
# ---------------------------------------------------------------------------

if __name__ == '__main__':

    log = logging.getLogger('grizzled.os')
    hdlr = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s', '%T')
    hdlr.setFormatter(formatter)
    log.addHandler(hdlr)
    log.setLevel(logging.DEBUG)

    log.debug('Before daemonizing, PID=%d' % _os.getpid())
    daemonize(no_close=True)
    log.debug('After daemonizing, PID=%d' % _os.getpid())
    log.debug('Daemon is sleeping for 10 seconds')

    import time
    time.sleep(10)

    log.debug('Daemon exiting')
    sys.exit(0)



#!/usr/bin/env python
#
# EasyInstall setup script for The Grizzled Utility Library
#
# $Id: 60ec5a51fd8b03d0a0746e5d4618c688b469e2c4 $
# ---------------------------------------------------------------------------

import ez_setup
ez_setup.use_setuptools(download_delay=2)
from setuptools import setup, find_packages
import sys
import os
import imp

here = os.path.dirname(os.path.abspath(__file__))
module_file = os.path.join(here, 'grizzled', '__init__.py')
module = imp.load_module('grizzled', open(module_file), module_file,
                         ('__init__.py', 'r', imp.PY_SOURCE))

# Now the setup stuff.

setup (name             = 'grizzled',
       version          = module.version,
       description      = module.title,
       long_description = module.__doc__,
       packages         = find_packages(),
       py_modules       = ['ez_setup'],
       url              = module.url,
       license          = module.version,
       author           = module.author,
       author_email     = module.email,
       test_suite       = 'nose.collector',
       classifiers = [
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ]
)



# $Id$

"""
Input/Output utility methods and classes.
"""

from __future__ import absolute_import

__docformat__ = "restructuredtext en"

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

import os
import zipfile

# ---------------------------------------------------------------------------
# Exports
# ---------------------------------------------------------------------------

__all__ = ['AutoFlush', 'MultiWriter', 'PushbackFile', 'Zip']

# ---------------------------------------------------------------------------
# Classes
# ---------------------------------------------------------------------------

class AutoFlush(object):
    """
    An ``AutoFlush`` wraps a file-like object and flushes the output
    (via a call to ``flush()`` after every write operation. Here's how
    to use an ``AutoFlush`` object to force standard output to flush after
    every write:

    .. python::

        import sys
        from grizzled.io import AutoFlush

        sys.stdout = AutoFlush(sys.stdout)
    """
    def __init__(self, f):
        """
        Create a new ``AutoFlush`` object to wrap a file-like object.

        :Parameters:
            f : file
                A file-like object that contains both a ``write()`` method
                and a ``flush()`` method.
        """
        self.__file = f

    def write(self, buf):
        """
        Write the specified buffer to the file.

        :Parameters:
            buf : str or bytes
                buffer to write
        """
        self.__file.write(buf)
        self.__file.flush()

    def flush(self):
        """
        Force a flush.
        """
        self.__file.flush()

    def truncate(self, size=-1):
        """
        Truncate the underlying file. Might fail.

        :Parameters:
            size : int
                Where to truncate. If less than 0, then file's current position
                is used.
        """
        if size < 0:
            size = self.__file.tell()
        self.__file.truncate(size)

    def tell(self):
        """
        Return the file's current position, if applicable.

        :rtype:  int
        :return: Current file position
        """
        return self.__file.tell()

    def seek(self, offset, whence=os.SEEK_SET):
        """
        Set the file's current position. The ``whence`` argument is optional;
        legal values are:

         - ``os.SEEK_SET`` or 0: absolute file positioning (default)
         - ``os.SEEK_CUR`` or 1: seek relative to the current position
         - ``os.SEEK_END`` or 2: seek relative to the file's end

        There is no return value. Note that if the file is opened for
        appending (mode 'a' or 'a+'), any ``seek()`` operations will be undone
        at the next write. If the file is only opened for writing in append
        mode (mode 'a'), this method is essentially a no-op, but it remains
        useful for files opened in append mode with reading enabled (mode
        'a+'). If the file is opened in text mode (without 'b'), only offsets
        returned by ``tell()`` are legal. Use of other offsets causes
        undefined behavior.

        Note that not all file objects are seekable.

        :Parameters:
            offset : int
                where to seek
            whence : int
                see above
        """
        self.__file.seek(offset, whence)

    def fileno(self):
        """
        Return the integer file descriptor used by the underlying file.

        :rtype:  int
        :return: the file descriptor
        """
        return self.__file.fileno()

class MultiWriter(object):
    """
    Wraps multiple file-like objects so that they all may be written at once.
    For example, the following code arranges to have anything written to
    ``sys.stdout`` go to ``sys.stdout`` and to a temporary file:

    .. python::

        import sys
        from grizzled.io import MultiWriter

        sys.stdout = MultiWriter(sys.__stdout__, open('/tmp/log', 'w'))
    """
    def __init__(self, *args):
        """
        Create a new ``MultiWriter`` object to wrap one or more file-like
        objects.

        :Parameters:
            args : iterable
                One or more file-like objects to wrap
        """
        self.__files = args

    def write(self, buf):
        """
        Write the specified buffer to the wrapped files.

        :Parameters:
            buf : str or bytes
                buffer to write
        """
        for f in self.__files:
            f.write(buf)

    def flush(self):
        """
        Force a flush.
        """
        for f in self.__files:
            f.flush()

    def close(self):
        """
        Close all contained files.
        """
        for f in self.__files:
            f.close()

class PushbackFile(object):
    """
    A file-like wrapper object that permits pushback.
    """
    def __init__(self, f):
        """
        Create a new ``PushbackFile`` object to wrap a file-like object.

        :Parameters:
            f : file
                A file-like object that contains both a ``write()`` method
                and a ``flush()`` method.
        """
        self.__buf = [c for c in ''.join(f.readlines())]

    def write(self, buf):
        """
        Write the specified buffer to the file. This method throws an
        unconditional exception, since ``PushbackFile`` objects are read-only.

        :Parameters:
            buf : str or bytes
                buffer to write

        :raise NotImplementedError: unconditionally
        """
        raise NotImplementedError, 'PushbackFile is read-only'

    def pushback(self, s):
        """
        Push a character or string back onto the input stream.

        :Parameters:
            s : str
                the string to push back onto the input stream
        """
        self.__buf = [c for c in s] + self.__buf

    unread=pushback

    def read(self, n=-1):
        """
        Read *n* bytes from the open file.

        :Parameters:
            n : int
                Number of bytes to read. A negative number instructs
                ``read()`` to read all remaining bytes.

        :return: the bytes read
        """
        resultBuf = None
        if n > len(self.__buf):
            n = len(self.__buf)

        if (n < 0) or (n >= len(self.__buf)):
            resultBuf = self.__buf
            self.__buf = []

        else:
            resultBuf = self.__buf[0:n]
            self.__buf = self.__buf[n:]

        return ''.join(resultBuf)

    def readline(self, length=-1):
        """
        Read the next line from the file.

        :Parameters:
            length : int
                a length hint, or negative if you don't care

        :rtype:  str
        :return: the line
        """
        i = 0
        while i < len(self.__buf) and (self.__buf[i] != '\n'):
            i += 1

        result = self.__buf[0:i+1]
        self.__buf = self.__buf[i+1:]
        return ''.join(result)

    def readlines(self, sizehint=0):
        """
        Read all remaining lines in the file.

        :rtype:  list
        :return: list of lines
        """
        return self.read(-1)

    def __iter__(self):
        return self

    def next(self):
        """A file object is its own iterator.

        :rtype: str
        :return: the next line from the file

        :raise StopIteration: end of file
        :raise IncludeError: on error
        """
        line = self.readline()
        if (line == None) or (len(line) == 0):
            raise StopIteration
        return line

    def close(self):
        """Close the file. A no-op in this class."""
        pass

    def flush(self):
        """
        Force a flush. This method throws an unconditional exception, since
        ``PushbackFile`` objects are read-only.

        :raise NotImplementedError: unconditionally
        """
        raise NotImplementedError, 'PushbackFile is read-only'

    def truncate(self, size=-1):
        """
        Truncate the underlying file.  This method throws an unconditional exception, since
        ``PushbackFile`` objects are read-only.

        :Parameters:
            size : int
                Where to truncate. If less than 0, then file's current
                position is used

        :raise NotImplementedError: unconditionally
        """
        raise NotImplementedError, 'PushbackFile is read-only'

    def tell(self):
        """
        Return the file's current position, if applicable. This method throws
        an unconditional exception, since ``PushbackFile`` objects are
        read-only.

        :rtype:  int
        :return: Current file position

        :raise NotImplementedError: unconditionally
        """
        raise NotImplementedError, 'PushbackFile is not seekable'

    def seek(self, offset, whence=os.SEEK_SET):
        """
        Set the file's current position. This method throws an unconditional
        exception, since ``PushbackFile`` objects are not seekable.

        :Parameters:
            offset : int
                where to seek
            whence : int
                see above

        :raise NotImplementedError: unconditionally
        """
        raise NotImplementedError, 'PushbackFile is not seekable'

    def fileno(self):
        """
        Return the integer file descriptor used by the underlying file.

        :rtype:  int
        :return: the file descriptor
        """
        return -1

class Zip(zipfile.ZipFile):
    """
    ``Zip`` extends the standard ``zipfile.ZipFile`` class and provides a
    method to extract the contents of a zip file into a directory. Adapted
    from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/252508.
    """
    def __init__(self, file, mode="r",
                 compression=zipfile.ZIP_STORED,
                 allow_zip64=False):
        """
        Constructor. Initialize a new zip file.

        :Parameters:
            file : str
                path to zip file
            mode : str
                open mode. Valid values are 'r' (read), 'w' (write), and
                'a' (append)
            compression : int
                Compression type. Valid values: ``zipfile.ZIP_STORED`,
                ``zipfile.ZIP_DEFLATED``
            allow_zip64 : bool
                Whether or not Zip64 extensions are to be used
        """
        zipfile.ZipFile.__init__(self, file, mode, compression, allow_zip64)
        self.zipFile = file

    def extract(self, output_dir):
        """
        Unpack the zip file into the specified output directory.

        :Parameters:
            output_dir : str
                path to output directory. The directory is
                created if it doesn't already exist.
        """
        if not output_dir.endswith(':') and not os.path.exists(output_dir):
            os.mkdir(output_dir)

        num_files = len(self.namelist())

        # extract files to directory structure
        for i, name in enumerate(self.namelist()):
            if not name.endswith('/'):
                directory = os.path.dirname(name)
                if directory == '':
                    directory = None
                if directory:
                    directory = os.path.join(output_dir, directory)
                    if not os.path.exists(directory):
                        os.makedirs(directory)

                outfile = open(os.path.join(output_dir, name), 'wb')
                outfile.write(self.read(name))
                outfile.flush()
                outfile.close()



'''
Introduction
============

Based on the standard Python ``ConfigParser`` module, this module provides an
enhanced configuration parser capabilities. ``Configuration`` is a drop-in
replacement for ``ConfigParser``.

A configuration file is broken into sections, and each section is
introduced by a section name in brackets. For example::

    [main]
    installationDirectory=/usr/local/foo
    programDirectory: /usr/local/foo/programs

    [search]
    searchCommand: find /usr/local/foo -type f -name "*.class"

    [display]
    searchFailedMessage=Search failed, sorry.


Section Name Syntax
===================

A section name can consist of alphabetics, numerics, underscores and
periods. There can be any amount of whitespace before and after the
brackets in a section name; the whitespace is ignored.

Variable Syntax
===============

Each section contains zero or more variable settings.

 - Similar to a Java ``Properties`` file, the variables are specified as
   name/value pairs, separated by an equal sign ("=") or a colon (":").
 - Variable names are case-sensitive and may contain alphabetics, numerics,
   underscores and periods (".").
 - Variable values may contain anything at all. Leading whitespace in the
   value is skipped. The way to include leading whitespace in a value is
   escape the whitespace characters with backslashes.

Variable Substitution
=====================

A variable value can interpolate the values of other variables, using a
variable substitution syntax. The general form of a variable reference is
``${section_name:var_name}``.

  - *section_name* is the name of the section containing the variable to
    substitute; if omitted, it defaults to the current section.
  - *var_name* is the name of the variable to substitute.

Default values
--------------

You can also specify a default value for a variable, using this syntax::

    ${foo?default}
    ${section:foo?default}

That is, the sequence ``?default`` after a variable name specifies the
default value if the variable has no value. (Normally, if a variable has
no value, it is replaced with an empty string.) Defaults can be useful,
for instance, to allow overrides from the environment. The following example
defines a log file directory that defaults to "/tmp", unless environment
variable LOGDIR is set to a non-empty value::

    logDirectory: ${env:LOGDIR?/var/log}

Special section names
---------------------

The section names "env", and "program" are reserved for special
pseudosections.

The ``env`` pseudosection
~~~~~~~~~~~~~~~~~~~~~~~~~

The "env" pseudosection is used to interpolate values from the environment. On
UNIX systems, for instance, ``${env:HOME}`` substitutes home directory of the
current user. On some versions of Windows, ``${env:USERNAME}`` will substitute
the name of the user.

Note: On UNIX systems, environment variable names are typically
case-sensitive; for instance, ``${env:USER}`` and ``${env:user}`` refer to
different environment variables. On Windows systems, environment variable
names are typically case-insensitive; ``${env:USERNAME}`` and
``${env:username}`` are equivalent.

The ``program`` pseudosection
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The "program" pseudosection is a placeholder for various special variables
provided by the Configuration class. Those variables are:

``cwd``
    The current working directory. Thus, ``${program:cwd}`` will substitute
    the working directory, using the appropriate system-specific
    file separator (e.g., "/" on Unix, "\\" on Windows).

``name``
    The calling program name. Equivalent to the Python expression
    ``os.path.basename(sys.argv[0])``

``now``
    The current time, formatted using the ``time.strftime()`` format
    ``"%Y-%m-%d %H:%M:%S"`` (e.g., "2008-03-03 16:15:27")

Includes
--------

A special include directive permits inline inclusion of another
configuration file. The include directive takes two forms::

   %include "path"
   %include "URL"

For example::

    %include "/home/bmc/mytools/common.cfg"
    %include "http://configs.example.com/mytools/common.cfg"

The included file may contain any content that is valid for this parser. It
may contain just variable definitions (i.e., the contents of a section,
without the section header), or it may contain a complete configuration
file, with individual sections.

Note: Attempting to include a file from itself, either directly or
indirectly, will cause the parser to throw an exception.

Replacing ``ConfigParser``
==========================

You can use this class anywhere you would use the standard Python
``ConfigParser`` class. Thus, to change a piece of code to use enhanced
configuration, you might change this:

.. python::

    import ConfigParser

    config = ConfigParser.SafeConfigParser()
    config.read(configPath)

to this:

.. python::

    from grizzled.config import Configuration

    config = Configuration()
    config.read(configPath)


Sometimes, however, you have to use an API that expects a path to a
configuration file that can *only* be parsed with the (unenhanced)
``ConfigParser`` class. In that case, you simply use the ``preprocess()``
method:

.. python::

    import logging
    from grizzled import config

    logging.config.fileConfig(config.preprocess(pathToConfig))

That will preprocess the enhanced configuration file, producing a file
that is suitable for parsing by the standard Python ``config`` module.
'''

from __future__ import absolute_import

__docformat__ = "restructuredtext en"

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

import ConfigParser
import logging
import string
import os
import time
import sys
import re

from grizzled.exception import ExceptionWithMessage
from grizzled.collections import OrderedDict

# ---------------------------------------------------------------------------
# Exports
# ---------------------------------------------------------------------------

__all__ = ['Configuration', 'preprocess',
           'NoOptionError', 'NoSectionError', 'NoVariableError']

# ---------------------------------------------------------------------------
# Globals
# ---------------------------------------------------------------------------

log = logging.getLogger('grizzled.config')
NoOptionError = ConfigParser.NoOptionError
NoSectionError = ConfigParser.NoSectionError

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

# Used with _ConfigDict

SECTION_OPTION_DELIM = r':'

# Section name pattern
SECTION_NAME_PATTERN = r'([_.a-zA-Z][_.a-zA-Z0-9]+)'

# Pattern of an identifier local to a section.
VARIABLE_NAME_PATTERN = r'([_a-zA-Z][_a-zA-Z0-9]+)(\?[^}]+)?'

# Pattern of an identifier matched by our version of string.Template.
# Intended to match:
#
#     ${section:option}         variable 'option' in section 'section'
#     ${section:option?default} variable 'option' in section 'section', default
#                               value 'default'
#     ${option}                 variable 'option' in the current section
#     ${option?default}         variable 'option' in the current section,
#                               default value 'default'
VARIABLE_REF_PATTERN = SECTION_NAME_PATTERN + SECTION_OPTION_DELIM +\
                       VARIABLE_NAME_PATTERN +\
                       r'|' +\
                       VARIABLE_NAME_PATTERN

# Simple variable reference
SIMPLE_VARIABLE_REF_PATTERN = r'\$\{' + VARIABLE_NAME_PATTERN + '\}'

# Special sections
ENV_SECTION = 'env'
PROGRAM_SECTION = 'program'

# ---------------------------------------------------------------------------
# Classes
# ---------------------------------------------------------------------------

class NoVariableError(ExceptionWithMessage):
    """
    Thrown when a configuration file attempts to substitute a nonexistent
    variable, and the ``Configuration`` object was instantiated with
    ``strict_substitution`` set to ``True``.
    """
    pass

class Configuration(ConfigParser.SafeConfigParser):
    """
    Configuration file parser. See the module documentation for details.
    """

    def __init__(self,
                 defaults=None,
                 permit_includes=True,
                 use_ordered_sections=False,
                 strict_substitution=False):
        """
        Construct a new ``Configuration`` object.

        :Parameters:
            defaults : dict
                dictionary of default values
            permit_includes : bool
                whether or not to permit includes
            use_ordered_sections : bool
                whether or not to use an ordered dictionary for the section
                names. If ``True``, then a call to ``sections()`` will return
                the sections in the order they were encountered in the file.
                If ``False``, the order is based on the hash keys for the
                sections' names.
            strict_substitution : bool
                If ``True``, then throw an exception if attempting to
                substitute a non-existent variable. Otherwise, simple
                substitute an empty value.
        """
        ConfigParser.SafeConfigParser.__init__(self, defaults)
        self.__permit_includes = permit_includes
        self.__use_ordered_sections = use_ordered_sections
        self.__strict_substitution = strict_substitution

        if use_ordered_sections:
            self._sections = OrderedDict()

    def defaults(self):
        """
        Returns the instance-wide defaults.

        :rtype:  dict
        :return: the instance-wide defaults, or ``None`` if there aren't any
        """
        return ConfigParser.SafeConfigParser.defaults(self)

    @property
    def sections(self):
        """
        Get the list of available sections, not including ``DEFAULT``. It's
        not really useful to call this method before calling ``read()`` or
        ``readfp()``.
        
        Returns a list of sections.
        """
        return ConfigParser.SafeConfigParser.sections(self)

    def add_section(self, section):
        """
        Add a section named *section* to the instance. If a section by the
        given name already exists, ``DuplicateSectionError`` is raised.

        :Parameters:
            section : str
                name of section to add

        :raise DuplicateSectionError: section already exists
        """
        ConfigParser.SafeConfigParser.add_section(self, section)

    def has_section(self, section):
        """
        Determine whether a section exists in the configuration. Ignores
        the ``DEFAULT`` section.

        :Parameters:
            section : str
                name of section

        :rtype:  bool
        :return: ``True`` if the section exists in the configuration, ``False``
                 if not.
        """
        return ConfigParser.SafeConfigParser.has_section(self, section)

    def options(self, section):
        """
        Get a list of options available in the specified section.

        :Parameters:
            section : str
                name of section

        :rtype:  list
        :return: list of available options. May be empty.

        :raise NoSectionError: no such section
        """
        return ConfigParser.SafeConfigParser.options(self, section)

    def has_option(self, section, option):
        """
        Determine whether a section has a specific option.

        :Parameters:
            section : str
                name of section
            option : str
                name of option to check

        :rtype:  bool
        :return: ``True`` if the section exists in the configuration and
                 has the specified option, ``False`` if not.
        """
        return ConfigParser.SafeConfigParser.has_option(self, section, option)

    def read(self, filenames):
        """
        Attempt to read and parse a list of filenames or URLs, returning a
        list of filenames or URLs which were successfully parsed. If
        *filenames* is a string or Unicode string, it is treated as a single
        filename or URL. If a file or URL named in filenames cannot be opened,
        that file will be ignored. This is designed so that you can specify a
        list of potential configuration file locations (for example, the
        current directory, the user's home directory, and some system-wide
        directory), and all existing configuration files in the list will be
        read. If none of the named files exist, the ``Configuration`` instance
        will contain an empty dataset. An application which requires initial
        values to be loaded from a file should load the required file or files
        using ``readfp()`` before calling ``read()`` for any optional files:

        .. python::

            import Configuration
            import os

            config = Configuration.Configuration()
            config.readfp(open('defaults.cfg'))
            config.read(['site.cfg', os.path.expanduser('~/.myapp.cfg')])

        :Parameters:
            filenames : list or string
                list of file names or URLs, or the string for a single filename
                or URL

        :rtype:  list
        :return: list of successfully parsed filenames or URLs
        """
        if isinstance(filenames, basestring):
            filenames = [filenames]

        newFilenames = []
        for filename in filenames:
            try:
                self.__preprocess(filename, filename)
                newFilenames += [filename]
            except IOError:
                log.exception('Error reading "%s"' % filename)

        return newFilenames

    def readfp(self, fp, filename=None):
        '''
        Read and parse configuration data from a file or file-like object.
        (Only the ``readline()`` moethod is used.)

        :Parameters:
            fp : file
                File-like object with a ``readline()`` method
            filename : str
                Name associated with ``fp``, for error messages. If omitted or
                ``None``, then ``fp.name`` is used. If ``fp`` has no ``name``
                attribute, then ``"<???">`` is used.
        '''
        self.__preprocess(fp, filename)

    def get(self, section, option, optional=False):
        """
        Get an option from a section.

        :Parameters:
            section : str
                name of section
            option : str
                name of option to check
            optional : bool
                ``True`` to return None if the option doesn't exist. ``False``
                to throw an exception if the option doesn't exist.

        :rtype:  str
        :return: the option value

        :raise NoSectionError: no such section
        :raise NoOptionError: no such option in the section
        """
        def do_get(section, option):
            val = ConfigParser.SafeConfigParser.get(self, section, option)
            if len(val.strip()) == 0:
                raise ConfigParser.NoOptionError(option, section)
            return val

        if optional:
            return self.__get_optional(do_get, section, option)
        else:
            return do_get(section, option)

    def getint(self, section, option, optional=False):
        """
        Convenience method that coerces the result of a call to
        ``get()`` to an ``int``.

        :Parameters:
            section : str
                name of section
            option : str
                name of option to check
            optional : bool
                ``True`` to return None if the option doesn't exist. ``False``
                to throw an exception if the option doesn't exist.

        :rtype:  int
        :return: the option value

        :raise NoSectionError: no such section
        :raise NoOptionError: no such option in the section
        """
        def do_get(section, option):
            return ConfigParser.SafeConfigParser.getint(self, section, option)

        if optional:
            return self.__get_optional(do_xget, section, option)
        else:
            return do_get(section, option)

    def getfloat(self, section, option, optional=False):
        """
        Convenience method that coerces the result of a call to ``get()`` to a
        ``float``.

        :Parameters:
            section : str
                name of section
            option : str
                name of option to check
            optional : bool
                ``True`` to return None if the option doesn't exist. ``False``
                to throw an exception if the option doesn't exist.

        :rtype:  float
        :return: the option value

        :raise NoSectionError: no such section
        :raise NoOptionError: no such option in the section
        """
        def do_get(section, option):
            return ConfigParser.SafeConfigParser.getfloat(self, section, option)

        if optional:
            return self.__get_optional(do_get, section, option)
        else:
            return do_get(section, option)

    def getboolean(self, section, option, optional=False):
        '''
        Convenience method that coerces the result of a call to ``get()`` to a
        boolean. Accepted boolean values are "1", "yes", "true", and "on",
        which cause this method to return True, and "0", "no", "false", and
        "off", which cause it to return False. These string values are checked
        in a case-insensitive manner. Any other value will cause it to raise
        ``ValueError``.

        :Parameters:a
            section : str
                name of section
            option : str
                name of option to check
            optional : bool
                ``True`` to return None if the option doesn't exist. ``False``
                to throw an exception if the option doesn't exist.

        :rtype:  bool
        :return: the option value (``True`` or ``False``)

        :raise NoSectionError: no such section
        :raise NoOptionError: no such option in the section
        :raise ValueError: non-boolean value encountered
        '''
        def do_get(section, option):
            return ConfigParser.SafeConfigParser.getboolean(self,
                                                            section,
                                                            option)

        if optional:
            return self.__get_optional(do_get, section, option)
        else:
            return do_get(section, option)

    def getlist(self, section, option, sep=None, optional=False):
        '''
        Convenience method that coerces the result of a call to ``get()`` to a
        list. The value is split using the separator(s) specified by the
        ``sep`` argument. A ``sep`` value of ``None`` uses white space. The
        result is a list of string values.

        :Parameters:
            section : str
                name of section
            option : str
                name of option to check
            sep : str
                list element separator to use. Defaults to white space.
            optional : bool
                ``True`` to return None if the option doesn't exist. ``False``
                to throw an exception if the option doesn't exist.

        :rtype:  bool
        :return: the option value (``True`` or ``False``)

        :raise NoSectionError: no such section
        :raise NoOptionError: no such option in the section
        '''
        def do_get(section, option):
            value = ConfigParser.SafeConfigParser.get(self, section, option)
            return value.split(sep)

        if optional:
            return self.__get_optional(do_get, section, option)
        else:
            return do_get(section, option)

    def get_one_of(self,
                   section,
                   options,
                   optional=False,
                   default=None,
                   value_type=str):
        '''
        Retrieve at most one of a list or set of options from a section. This
        method is useful if there are multiple possible names for a single
        option. For example, suppose you permit either a ``user_name`` or a
        ``login_name`` option, but not both, in a section called
        ``credentials``. You can use the following code to retrieve the
        option value:

        .. python::

            from grizzled.config import Configuration

            config = Configuration()
            config.read('/path/to/config')
            user = config.get_one_of('credentials', ['user_name', 'login_name'])

        If both options exist, ``get_one_of()`` will a ``NoOptionError``. If
        neither option exists, ``get_one_of()`` will throw a ``NoOptionError``
        if ``optional`` is ``False`` and there's no default value; otherwise,
        it will return the default value.

        :Parameters:
            section : str
                name of section
            options : list or set
                list or set of allowable option names
            optional : bool
                ``True`` to return None if the option doesn't exist. ``False``
                to throw an exception if the option doesn't exist.
            default : str
                The default value, if the option does not exist.
            value_type : type
                The type to which to coerce the value. The value is coerced by
                casting.

        :rtype:  str
        :return: the option value, or ``None`` if nonexistent and
                 ``optional`` is ``True``

        :raise NoSectionError: no such section
        :raise NoOptionError: none of the named options are in the section
        '''
        value = None
        if value_type is bool:
            get = self.getboolean
        else:
            get = self.get

        for option in options:
            value = get(section, option, optional=True)
            if value:
                break

        if value is None:
            value = default

        if (value is None) and (not optional):
            raise NoOptionError('Section "%s" must contain exactly one of the '
                                'following options: %s' %
                                (section, ', '.join(list(options))))

        if value is not None:
            if not (value_type in (bool, str)):
                value = eval('%s(%s)' % (value_type.__name__, value))

        return value

    def items(self, section):
        """
        Get all items in a section.

        :Parameters:
            section : str
                name of section

        :rtype:  list
        :return: a list of (*name*, *value*) tuples for each option in
                 in *section*

        :raise NoSectionError: no such section
        """
        return ConfigParser.SafeConfigParser.items(self, section)

    def set(self, section, option, value):
        """
        If the given section exists, set the given option to the specified
        value; otherwise raise ``NoSectionError``.

        :Parameters:
            section : str
                name of section
            option : str
                name of option to check
            value : str
                the value to set

        :raise NoSectionError: no such section
        """
        ConfigParser.SafeConfigParser.set(self, section, option, value)

    def write(self, fileobj):
        """
        Write a representation of the configuration to the specified file-like
        object. This output can be parsed by a future ``read()`` call.

        NOTE: Includes and variable references are ``not`` reconstructed.
        That is, the configuration data is written in *expanded* form.

        :Parameters:
            fileobj : file
                file-like object to which to write the configuration
        """
        ConfigParser.SafeConfigParser.write(self, fileobj)

    def remove_section(self, section):
        """
        Remove a section from the instance. If a section by the given name
        does not exist, ``NoSectionError`` is raised.

        :Parameters:
            section : str
                name of section to remove

        :raise NoSectionError: no such section
        """
        ConfigParser.SafeConfigParser.remove_section(self, section)

    def optionxform(self, option_name):
        """
        Transforms the option name in ``option_name`` as found in an input
        file or as passed in by client code to the form that should be used in
        the internal structures. The default implementation returns a
        lower-case version of ``option_name``; subclasses may override this or
        client code can set an attribute of this name on instances to affect
        this behavior. Setting this to ``str()``, for example, would make
        option names case sensitive.
        """
        return option_name.lower()

    def __get_optional(self, func, section, option):
        try:
            return func(section, option)
        except ConfigParser.NoOptionError:
            return None
        except ConfigParser.NoSectionError:
            return None

    def __preprocess(self, fp, name):

        try:
            fp.name
        except AttributeError:
            try:
                fp.name = name
            except TypeError:
                # Read-only. Oh, well.
                pass
            except AttributeError:
                # Read-only. Oh, well.
                pass

        if self.__permit_includes:
            # Preprocess includes.
            from grizzled.file import includer
            tempFile = includer.preprocess(fp)
            fp = tempFile

        # Parse the resulting file into a local ConfigParser instance.

        parsedConfig = ConfigParser.SafeConfigParser()

        if self.__use_ordered_sections:
            parsedConfig._sections = OrderedDict()

        parsedConfig.optionxform = str
        parsedConfig.read(fp)

        # Process the variable substitutions.

        self.__normalizeVariableReferences(parsedConfig)
        self.__substituteVariables(parsedConfig)

    def __normalizeVariableReferences(self, sourceConfig):
        """
        Convert all section-local variable references (i.e., those that don't
        specify a section) to fully-qualified references. Necessary for
        recursive references to work.
        """
        simpleVarRefRe = re.compile(SIMPLE_VARIABLE_REF_PATTERN)
        for section in sourceConfig.sections():
            for option in sourceConfig.options(section):
                value = sourceConfig.get(section, option, raw=True)
                oldValue = value
                match = simpleVarRefRe.search(value)
                while match:
                    value = value[0:match.start(1)] +\
                            section +\
                            SECTION_OPTION_DELIM +\
                            value[match.start(1):]
                    match = simpleVarRefRe.search(value)

                sourceConfig.set(section, option, value)

    def __substituteVariables(self, sourceConfig):
        mapping = _ConfigDict(sourceConfig, self.__strict_substitution)
        for section in sourceConfig.sections():
            mapping.section = section
            self.add_section(section)
            for option in sourceConfig.options(section):
                value = sourceConfig.get(section, option, raw=True)

                # Repeatedly substitute, to permit recursive references

                previousValue = ''
                while value != previousValue:
                    previousValue = value
                    value = _ConfigTemplate(value).safe_substitute(mapping)

                self.set(section, option, value)

class _ConfigTemplate(string.Template):
    """
    Subclass of string.Template that handles our configuration variable
    reference syntax.
    """
    idpattern = VARIABLE_REF_PATTERN

class _ConfigDict(dict):
    """
    Dictionary that knows how to dereference variables within a parsed config.
    Only used internally.
    """
    idPattern = re.compile(VARIABLE_REF_PATTERN)
    def __init__(self, parsedConfig, strict_substitution):
        self.__config = parsedConfig
        self.__strict_substitution = strict_substitution
        self.section = None

    def __getitem__(self, key):
        try:
            # Match against the ID regular expression. (If the match fails,
            # it's a bug, since we shouldn't be in here unless it does.)

            match = self.idPattern.search(key)
            assert(match)

            # Now, get the value.

            default = None
            if SECTION_OPTION_DELIM in key:
                if match.group(3):
                    default = self.__extract_default(match.group(3))

                section = match.group(1)
                option = match.group(2)
            else:
                section = self.section
                default = None
                option = match.group(3)
                if match.group(4):
                    default = self.__extract_default(match.group(3))

            result = self.__value_from_section(section, option)

        except KeyError:
            result = default

        except ConfigParser.NoSectionError:
            result = default

        except ConfigParser.NoOptionError:
            result = default

        if not result:
            if self.__strict_substitution:
                raise NoVariableError, 'No such variable: "%s"' % key
            else:
                result = ''

        return result

    def __extract_default(self, s):
        default = s
        if default:
            default = default[1:]  # strip leading '?'
            if len(default) == 0:
                default = None

        return default

    def __value_from_program_section(self, option):
        return {
            'cwd'  : os.getcwd(),
            'now'  : time.strftime('%Y-%m-%d %H:%M:%S'),
            'name' : os.path.basename(sys.argv[0])
               }[option]

    def __value_from_section(self, section, option):
        result = None
        if section == 'env':
            result = os.environ[option]
            if len(result) == 0:
                raise KeyError, option

        elif section == 'program':
            result = self.__value_from_program_section(option)

        else:
            result = self.__config.get(section, option)

        return result

# ---------------------------------------------------------------------------
# Functions
# ---------------------------------------------------------------------------

def preprocess(file_or_url, defaults=None):
    """
    This function preprocesses a file or URL for a configuration file,
    processing all includes and substituting all variables. It writes a new
    configuration file to a temporary file (or specified output file). The
    new configuration file can be read by a standard ``ConfigParser``
    object. Thus, this method is useful when you have an extended
    configuration file that must be passed to a function or object that can
    only read a standard ``ConfigParser`` file.

    For example, here's how you might use the Python ``logging`` API with an
    extended configuration file:

    .. python::

        from grizzled.config import Configuration
        import logging

        logging.config.fileConfig(Configuration.preprocess('/path/to/config')

    :Parameters:
        file_or_url :  str
           file or URL to read and preprocess
        defaults : dict
            defaults to pass through to the config parser

    :rtype: string
    :return: Path to a temporary file containing the expanded configuration.
             The file will be deleted when the program exits, though the caller
             is free to delete it sooner.
    """
    import tempfile
    import atexit

    def unlink(path):
        try:
            os.unlink(path)
        except:
            pass

    parser = Configuration(use_ordered_sections=True)
    parser.read(file_or_url)
    fd, path = tempfile.mkstemp(suffix='.cfg')
    atexit.register(unlink, path)
    parser.write(os.fdopen(fd, "w"))
    return path


# ---------------------------------------------------------------------------
# Main program (for testing)
# ---------------------------------------------------------------------------

if __name__ == '__main__':
    import sys

    format = '%(asctime)s %(name)s %(levelname)s %(message)s'
    logging.basicConfig(level=logging.DEBUG, format=format)

    configFile = sys.argv[1]
    config = Configuration()
    config.read(configFile)

    if len(sys.argv) > 2:
        for var in sys.argv[2:]:
            (section, option) = var.split(':')
            val = config.get(section, option, optional=True)
            print '%s=%s' % (var, val)
    else:
        config.write(sys.stdout)



# $Id: 87faaea2d73732a6c2e0e47cc657debc29f8268c $
# ---------------------------------------------------------------------------
 
"""
A dummy database driver, useful for testing.
"""

__docformat__ = "restructuredtext en"

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

from grizzled.db.base import DBDriver

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

BINARY = 0
NUMBER = 1
STRING = 2
DATETIME = 3
ROWID = 4

# ---------------------------------------------------------------------------
# Classes
# ---------------------------------------------------------------------------

class DummyCursor(object):
    def close(self):
        pass

    def execute(self, statement, parameters=None):
        self.rowcount = 0
        self.description = ""
        return None

    def fetchone(self):
        raise ValueError, "No results"

    def fetchall(self):
        raise ValueError, "No results"

    def fetchmany(self, n):
        raise ValueError, "No results"

class DummyDB(object):

    def __init__(self):
        pass

    def cursor(self):
        return DummyCursor()

    def commit(self):
        pass

    def rollback(self):
        pass

    def close(self):
        pass

class DummyDriver(DBDriver):
    """Dummy database driver, for testing."""

    def get_import(self):
        import dummydb
        return dummydb

    def get_display_name(self):
        return "Dummy"

    def do_connect(self,
                   host="localhost",
                   port=None,
                   user='',
                   password='',
                   database='default'):
        return dummydb.DummyDB()



"""
This module contains various Python decorators.
"""

__docformat__ = "restructuredtext en"

# ---------------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# Exports
# ---------------------------------------------------------------------------

__all__ = ['deprecated', 'abstract']

# ---------------------------------------------------------------------------
# Decorators
# ---------------------------------------------------------------------------

def deprecated(since=None, message=None):
    """
    Decorator for marking a function deprecated. Generates a warning on
    standard output if the function is called.

    Usage:
    
    .. python::

        from grizzled.decorators import deprecated

        class MyClass(object):

            @deprecated()
            def oldMethod(self):
                pass

    Given the above declaration, the following code will cause a
    warning to be printed (though the method call will otherwise succeed):
    
    .. python::

        obj = MyClass()
        obj.oldMethod()

    You may also specify a ``since`` argument, used to display a deprecation
    message with a version stamp (e.g., 'deprecated since ...'):
    
    .. python::

        from grizzled.decorators import deprecated

        class MyClass(object):

            @deprecated(since='1.2')
            def oldMethod(self):
                pass

    :Parameters:
        since : str
            version stamp, or ``None`` for none
        message : str
            optional additional message to print
    """
    def decorator(func):
        if since is None:
            buf = 'Method %s is deprecated.' % func.__name__
        else:
            buf = 'Method %s has been deprecated since version %s.' %\
                  (func.__name__, since)

        if message:
            buf += ' ' + message

        def wrapper(*__args, **__kw):
            import warnings
            warnings.warn(buf, category=DeprecationWarning, stacklevel=2)
            return func(*__args,**__kw)

        wrapper.__name__ = func.__name__
        wrapper.__dict__ = func.__dict__
        wrapper.__doc__ = func.__doc__
        return wrapper

    return decorator

def abstract(func):
    """
    Decorator for marking a method abstract. Throws a ``NotImplementedError``
    if an abstract method is called.

    Usage:
    
    .. python::

        from grizzled.decorators import abstract

        class MyAbstractClass(object):

            @abstract
            def abstractMethod(self):
                pass

        class NotReallyConcrete(MyAbstractClass):
            # Class doesn't define abstractMethod().

    Given the above declaration, the following code will cause an
    ``NotImplementedError``:
    
    .. python::

        obj = NotReallyConcrete()
        obj.abstractMethod()
    """
    def wrapper(*__args, **__kw):
        raise NotImplementedError('Missing required %s() method' %\
                                  func.__name__)
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper

def unimplemented(func):
    """
    Decorator for marking a function or method unimplemented. Throws a
    ``NotImplementedError`` if called. Note that this decorator is
    conceptually different from ``@abstract``. With ``@abstract``, the method
    is intended to be implemented by a subclass. With ``@unimplemented``, the
    method should never be implemented.

    Usage:
    
    .. python::

        from grizzled.decorators import unimplemented

        class ReadOnlyDict(dict):

            @unimplemented
            def __setitem__(self, key, value):
                pass
    """
    def wrapper(*__args, **__kw):
        raise NotImplementedError('Method or function "%s" is not implemented',
                                  func.__name__)
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper


# ---------------------------------------------------------------------------
# Main program, for testing
# ---------------------------------------------------------------------------

if __name__ == '__main__':
    @deprecated()
    def func1(a):
        pass

    @deprecated(since='1.2')
    def func2():
        pass

    func1(100)
    func2()

    class Foo(object):
        @abstract
        def foo(self):
            pass

    class Bar(Foo):
        pass

    b = Bar()
    try:
        b.foo()
        assert False
    except NotImplementedError, ex:
        print ex.message
